import os
import threading
from enum import Enum
from typing import Callable, Any, Optional

from sirius.constants import EnvironmentVariable
from sirius.exceptions import ApplicationException


class Environment(Enum):
    Production: str = "Production"
    Staging: str = "Staging"
    Development: str = "Development"


def get_environmental_variable(environmental_variable: EnvironmentVariable) -> str:
    value: Optional[str] = os.getenv(environmental_variable.value)
    if value is None:
        raise ApplicationException(f"Environment variable with the key is not available: {environmental_variable.value}")

    return value


def get_environment() -> Environment:
    environment: Optional[str] = os.getenv(EnvironmentVariable.ENVIRONMENT.value)
    try:
        return Environment.Development if environment is None else Environment(environment)
    except ValueError:
        raise ApplicationException(f"Invalid environment variable setup: {environment}")


def is_production_environment() -> bool:
    return Environment.Production == get_environment()


def is_staging_environment() -> bool:
    return Environment.Staging == get_environment()


def is_development_environment() -> bool:
    return Environment.Development == get_environment()


def threaded(func: Callable) -> Callable:
    def wrapper(*args: Any, **kwargs: Any) -> threading.Thread:
        thread: threading.Thread = threading.Thread(target=func, args=args, kwargs=kwargs)
        thread.start()
        return thread

    return wrapper



import json
from abc import abstractmethod
from dataclasses import dataclass
from typing import Dict, Any, List, Optional

import httpx
from httpx import Response, Cookies, Headers
from pydantic import BaseModel

from sirius import application_performance_monitoring
from sirius.application_performance_monitoring import Operation
from sirius.http_requests.exceptions import ClientSideException, ServerSideException


@dataclass
class HTTPResponse:
    response: Response
    response_code: int
    is_successful: bool
    headers: Headers
    data: Optional[Dict[Any, Any]] = None
    response_text: Optional[str] = None
    cookies: Optional[Cookies] = None

    def __init__(self, response: Response, *args: List[Any], **kwargs: Dict[str, Any]) -> None:
        self.response = response
        self.response_code = self.response.status_code
        self.is_successful = 200 <= self.response_code < 300
        self.response_text = self.response.text
        self.headers = response.headers
        self.cookies = response.cookies

        if self.is_successful and self.response_text is not None and self.response_text != "":
            self.data = self.response.json()

        super().__init__(*args, **kwargs)


class HTTPRequest:

    def __init__(self) -> None:
        pass

    @staticmethod
    def raise_http_exception(http_response: HTTPResponse) -> None:
        error_message: str = f"HTTP Exception\n" \
                             f"URL: {str(http_response.response.url)}\n" \
                             f"Headers: {str(http_response.headers)}\n" \
                             f"Method: {http_response.response.request.method.upper()}\n" \
                             f"Response Code: {http_response.response_code}\n" \
                             f"Response Text: {http_response.response_text}"

        if 400 <= http_response.response_code < 500:
            raise ClientSideException(error_message)
        else:
            raise ServerSideException(error_message)

    @application_performance_monitoring.transaction(Operation.HTTP_REQUEST, "GET")
    async def get(self, url: str, query_params: Optional[Dict[str, Any]] = None, headers: Optional[Dict[str, Any]] = None) -> HTTPResponse:
        async with httpx.AsyncClient() as client:
            http_response: HTTPResponse = HTTPResponse(await client.get(url, params=query_params, headers=headers))
            if not http_response.is_successful:
                HTTPRequest.raise_http_exception(http_response)

        return http_response

    @application_performance_monitoring.transaction(Operation.HTTP_REQUEST, "PUT")
    async def put(self, url: str, data: Dict[str, Any], headers: Optional[Dict[str, Any]] = None) -> HTTPResponse:
        async with httpx.AsyncClient() as client:
            http_response: HTTPResponse = HTTPResponse(await client.put(url, data=data, headers=headers))
            if not http_response.is_successful:
                HTTPRequest.raise_http_exception(http_response)

        return http_response

    @application_performance_monitoring.transaction(Operation.HTTP_REQUEST, "POST")
    async def post(self, url: str, data: Optional[Dict[str, Any]] = None, headers: Optional[Dict[str, Any]] = None) -> HTTPResponse:
        data_string: Optional[str] = None
        if headers is not None:
            headers["content-type"] = "application/json"

        if data is not None:
            data_string = json.dumps(data)

        async with httpx.AsyncClient() as client:
            http_response: HTTPResponse = HTTPResponse(await client.post(url, data=data_string, headers=headers))  # type: ignore[arg-type]
            if not http_response.is_successful:
                HTTPRequest.raise_http_exception(http_response)

        return http_response

    @application_performance_monitoring.transaction(Operation.HTTP_REQUEST, "DELETE")
    async def delete(self, url: str, headers: Optional[Dict[str, Any]] = None) -> HTTPResponse:
        async with httpx.AsyncClient() as client:
            http_response: HTTPResponse = HTTPResponse(await client.delete(url, headers=headers))
            if not http_response.is_successful:
                HTTPRequest.raise_http_exception(http_response)

        return http_response


class HTTPModel(BaseModel):
    _headers: Optional[Dict[str, Any]] = None

    @abstractmethod
    def __init__(self, **data: Any):
        super().__init__(**data)

    @classmethod
    async def get_one(cls, clazz: "HTTPModel", url: str, query_params: Optional[Dict[str, Any]] = None) -> "HTTPModel":
        response: HTTPResponse = await HTTPRequest().get(url=url, query_params=query_params, headers=clazz._headers)
        return clazz(**response.data)  # type: ignore[operator]

    @classmethod
    async def get_multiple(cls, clazz: "HTTPModel", url: str, query_params: Optional[Dict[str, Any]] = None) -> List["HTTPModel"]:
        response: HTTPResponse = await HTTPRequest().get(url=url, query_params=query_params, headers=clazz._headers)
        return [clazz(**data) for data in response.data]  # type: ignore[operator, union-attr]

    @classmethod
    async def post_return_one(cls, clazz: "HTTPModel", url: str, data: Optional[Dict[Any, Any]] = None) -> "HTTPModel":
        response: HTTPResponse = await HTTPRequest().post(url=url, data=data, headers=clazz._headers)
        return clazz(**response.data)  # type: ignore[operator]



import functools
from logging import Logger
from typing import Optional, Dict, Any, List, Callable

import beanie.odm.fields
from beanie import Document, init_beanie
from motor.motor_asyncio import AsyncIOMotorClient

from sirius import common, application_performance_monitoring
from sirius.application_performance_monitoring import Operation
from sirius.constants import EnvironmentVariable
from sirius.database.exceptions import DatabaseException, NonUniqueResultException, UncommittedRelationalDocumentException

client: AsyncIOMotorClient = None
database_name: str = common.get_environmental_variable(EnvironmentVariable.DATABASE_NAME)
logger: Logger = application_performance_monitoring.get_logger()


@application_performance_monitoring.transaction(Operation.DATABASE, f"Connect to the {database_name} Database")
async def initialize() -> None:
    global client, database_name
    if client is not None:
        return

    client = AsyncIOMotorClient(common.get_environmental_variable(EnvironmentVariable.MONGO_DB_CONNECTION_STRING))
    await init_beanie(database=client[database_name], document_models=DatabaseDocument.__subclasses__())


def transaction(transaction_name: str) -> Callable:
    def decorator(function: Callable) -> Callable:
        @application_performance_monitoring.transaction(Operation.AORTA_SIRIUS, transaction_name)
        @functools.wraps(function)
        async def wrapper(*args: List[Any], **kwargs: Dict[str, Any]) -> Any:
            await initialize()
            return await function(*args, **kwargs)

        return wrapper

    return decorator


class DatabaseDocument(Document):
    id: Optional[beanie.PydanticObjectId] = None

    class Settings:
        validate_on_save = True
        use_state_management = True
        state_management_save_previous = True

        def __init__(self) -> None:
            pass

    def __init__(self, *args: List[Any], **kwargs: Dict[Any, Any]) -> None:
        for key, value in kwargs.items():
            if isinstance(value, DatabaseDocument):
                if value.id is None:
                    raise UncommittedRelationalDocumentException(f"Uncommitted document is trying to be related\nUncommitted Object: {str(object)}\nObject being related: {str(self)}")
                kwargs[key] = value.id  # type: ignore[assignment]

        super().__init__(*args, **kwargs)  # type: ignore[no-untyped-call]

    @application_performance_monitoring.transaction(Operation.DATABASE, "Save")
    async def commit(self) -> "DatabaseDocument":
        if self.id is None:
            return await super().save(link_rule=beanie.WriteRules.WRITE)
        else:
            return await super().save_changes()

    @application_performance_monitoring.transaction(Operation.DATABASE, "Delete")
    async def remove(self) -> None:
        return await super().delete()

    @classmethod
    @application_performance_monitoring.transaction(Operation.DATABASE, "Find Single")
    async def find_unique(cls, *args: List[Any], fetch_links: bool = False) -> Optional["DatabaseDocument"]:
        results_list: List[DatabaseDocument] = await super().find_many(*args).to_list()  # type: ignore[call-overload]

        if len(results_list) == 1:
            result: DatabaseDocument = results_list[0]
        elif len(results_list) == 0:
            return None
        else:
            raise NonUniqueResultException(f"Non-unique result found\nCollection: {cls.__name__}\nSearch Criteria: {str(*args)}")

        if fetch_links:
            await result.fetch_all_links()  # type: ignore[no-untyped-call]
        return result

    @classmethod
    @application_performance_monitoring.transaction(Operation.DATABASE, "Find Multiple")
    async def find_multiple(cls, search_criteria: Dict[Any, Any], fetch_links: bool = False) -> List["DatabaseDocument"]:
        results_list: List[DatabaseDocument] = await super().find_many(search_criteria).to_list()
        if fetch_links:
            for result in results_list:
                await result.fetch_all_links()  # type: ignore[no-untyped-call]
        return results_list

    @staticmethod
    @application_performance_monitoring.transaction(Operation.DATABASE, "Save All")
    async def commit_all(database_document_list: List["DatabaseDocument"]) -> List["DatabaseDocument"]:
        return [await database_document.commit() for database_document in database_document_list]



import os
from typing import List

import requests
from requests import Response
from setuptools import setup, find_packages


def get_package_name() -> str:
    return "aorta_sirius" if is_production_branch() else "aorta_sirius_dev"


def get_next_version() -> str:
    response: Response = requests.get(f"https://pypi.org/pypi/{get_package_name()}/json")
    assert response.status_code == 200
    version_number: str = response.json()["info"]["version"]
    next_subversion_number: int = int(version_number.split(".")[-1]) + 1
    return ".".join(version_number.split(".")[0:-1], ) + "." + str(next_subversion_number)


def is_production_branch() -> bool:
    return "production" == os.getenv("GITHUB_REF_NAME")


def get_required_packages() -> List[str]:
    with open("requirements.txt", "r") as requirements_file:
        return requirements_file.read().split("\n")


setup(
    name=get_package_name(),
    version=get_next_version(),
    url="https://github.com/kontinuum-investments/Aorta-Sirius",
    author="Kavindu Athaudha",
    author_email="kavindu@kih.com.sg",
    packages=find_packages(where="src", include=["sirius*"]),
    package_dir={"": "src"},
    install_requires=get_required_packages()
)
